@@ -40,6 +40,14 @@ def pop_netstring(data):
         raise ValueError("not a netstring: %r" % (data,))
     return (body[:size],body[size+1:])
 
+def encode_netstring(data):
+    """Encode the given data as a netstring."""
+    return "%d:%s," % (len(data),data,)
+
+def encode_netstrings(datas):
+    """Encode the given sequence of data strings as joined netstrings."""
+    return "".join(encode_netstring(d) for d in datas)
+
 
 _QUOTED_SLASH_RE = re.compile("(?i)%2F")
 
